from .acc_accuracy_ratio import ACC_ACCURACY_RATE_CASE
from .acc_hit_ratio import ACC_HIT_RATIO_CASE
from .acc_ts import ACC_TS_SCORE_CASE
from .acc_miss_ratio import ACC_MISS_RATIO_CASE
from .acc_ets import ACC_ETS_SCORE_CASE
from .acc_bias_score import ACC_BIAS_SCORE_CASE
from .acc_false_alarm_ratio import ACC_FALSE_ALARM_RATE_CASE, ACC_FALSE_ALARM_RATIO_CASE



from typing import Union
from numbers import Number
from collections import Counter

import numpy as np
from scipy import stats

from ..utils import (
    convert_to_ndarray,
    fix_zero_division,
    assert_length,
    drop_nan,
)
from .binarize import threshold_binarize


@assert_length
@drop_nan
def calc_binary_quadrant_values(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate binary quadrant values

    Quadrant values includes:
        * hits
        * misses
        * false_alarms
        * correct_rejects

    Args:
        observation (ndarray): Binarized obervation data array
                               that consist of True and False
        forecast (ndarray): Binarized forecast data array
                            that consist of True and False

    Returns:
        tuple: (hits, misses, false_alarms, correct_rejects, total)
    """
    hits = Counter(observation & forecast)[True]
    misses = Counter((observation ^ forecast) & observation)[True]
    false_alarms = Counter((observation ^ forecast) & forecast)[True]
    correct_rejects = Counter(~observation & ~forecast)[True]
    total = hits + misses + false_alarms + correct_rejects

    assert len(observation) == total

    return hits, misses, false_alarms, correct_rejects, total


@assert_length
@fix_zero_division
@drop_nan
def calc_binary_accuracy_ratio(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate accuracy of binarized observation and forecast.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: The accuracy(%) of binarized forecast.
    """
    hits, _, _, correct_rejects, total = calc_binary_quadrant_values(
        observation, forecast
    )

    return ((hits + correct_rejects) / total) * 100


@assert_length
@fix_zero_division
@drop_nan
def calc_hit_ratio(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate hit ratio.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: Missing ratio(%)
    """
    hits, misses, _, correct_rejects, _ = calc_binary_quadrant_values(
        observation, forecast
    )

    return (hits / (hits + misses)) * 100


@assert_length
@fix_zero_division
@drop_nan
def calc_miss_ratio(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate missing ratio.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: Missing ratio(%)
    """
    hits, misses, _, correct_rejects, _ = calc_binary_quadrant_values(
        observation, forecast
    )

    return (misses / (hits + misses)) * 100


@assert_length
@fix_zero_division
@drop_nan
def calc_false_alarm_ratio(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate false alarm ratio, commonly known as FAR.

    Note that false alarm rate(POFD) is different from false alarm ratio(FAR).
    You can get more information about this from https://doi.org/10.1175/2009WAF2222300.1

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: False alarm ratio(%)
    """
    hits, _, false_alarms, _, _ = calc_binary_quadrant_values(observation, forecast)

    return (false_alarms / (hits + false_alarms)) * 100


@assert_length
@fix_zero_division
@drop_nan
def calc_false_alarm_rate(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate false alarm rate, commonly known as POFD.

    Note that false alarm rate(POFD) is different from false alarm ratio(FAR).
    You can get more information about this from https://doi.org/10.1175/2009WAF2222300.1

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: False alarm rate(%)
    """
    _, _, false_alarms, correct_rejects, _ = calc_binary_quadrant_values(
        observation, forecast
    )

    return (false_alarms / (correct_rejects + false_alarms)) * 100


@assert_length
@fix_zero_division
@drop_nan
def calc_ts(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate Threat score(TS)

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: Threat score(TS)
    """
    hits, misses, false_alarms, _, _ = calc_binary_quadrant_values(
        observation, forecast
    )

    return hits / (hits + false_alarms + misses)


@assert_length
@fix_zero_division
@drop_nan
def calc_ets(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate Equitable threat score(ETS)

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: Equitable threat score(ETS)
    """
    hits, misses, false_alarms, _, total = calc_binary_quadrant_values(
        observation, forecast
    )

    hits_ref = ((hits + misses) * (hits + false_alarms)) / total

    if hits - hits_ref == hits + false_alarms + misses - hits_ref:
        return 1.0
    return (hits - hits_ref) / (hits + false_alarms + misses - hits_ref)


@assert_length
@fix_zero_division
@drop_nan
def calc_bias_score(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate Bias score

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: Bias score
    """
    hits, misses, false_alarms, _, _ = calc_binary_quadrant_values(
        observation, forecast
    )

    return (hits + false_alarms) / (hits + misses)


@assert_length
@fix_zero_division
@convert_to_ndarray
@drop_nan
def calc_differ_accuracy_ratio(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    limit: Union[float, int] = 1,
) -> float:
    """Calculate the accuracy within a limited difference scope.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.
        limit (Union[float, int]): The limited(or tolerable) scope between
                                   observation and forecast. Default to 1.

    Returns:
        float: The difference accuracy(%).
    """

    diff = np.abs(observation - forecast)

    bins = np.full_like(diff, False)
    bins[diff <= limit] = True

    total = len(bins)
    counter = Counter(bins)

    return counter[True] / total * 100


@assert_length
@fix_zero_division
@convert_to_ndarray
@drop_nan
def calc_rmse(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate RMSE(root-mean-square error).

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

    Returns:
        float: The RMSE of forecast to observation.
    """
    return np.sqrt(np.sum((observation - forecast) ** 2) / len(observation))


@assert_length
@fix_zero_division
@convert_to_ndarray
@drop_nan
def calc_mae(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate MAE(mean absolute error).

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

    Returns:
        float: The MAE of forecast to observation.
    """
    return np.sum(np.abs(observation - forecast)) / len(observation)


@assert_length
@fix_zero_division
@convert_to_ndarray
@drop_nan
def calc_rss(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate the residual sum of square between forecast and observation.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

    Returns:
        float: The residual sum of square of forecast and observation.
    """
    return np.sum((observation - forecast) ** 2)


@assert_length
@fix_zero_division
@convert_to_ndarray
@drop_nan
def calc_chi_square(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate the χ2 between observation and forecast.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

    Returns:
        float: The chi square of forecast and observation.
    """
    return np.sum((observation - forecast) ** 2) / len(observation)


@assert_length
@convert_to_ndarray
@drop_nan
def calc_linregress(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> tuple:
    """Calculste the parameters of linear regression.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

    Returns:
        tuple: Aguments of linear regression as follow:
               slope, intercept, r_value, p_value, std_err
    """
    linreg_result = stats.linregress(forecast, observation)

    return tuple(round(p, 4) for p in linreg_result)


@assert_length
@drop_nan
def calc_threshold_accuracy_ratio(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    threshold: Number,
    compare: str = ">=",
) -> float:
    """Calculate accuracy ratio of forecast filtered by threshold.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.
        threshold (Number): The threshold to filter obervation and forecast.
                            This parameter should be used with `compare` parameter,
                            The `compare` parameter will control the logical operator.
        compare (str, optional): The logical operator applying `threshold` parameter.
                                * '>' greater
                                * '<' less
                                * '>=' greater or equal
                                * '<=' less or equal
                                Defaults to '>='.

    Returns:
        float: The accuracy ratio of forecast filtered by threshold.
    """
    obs_bins, fct_bins = threshold_binarize(
        observation, forecast, threshold=threshold, compare=compare
    )

    return calc_binary_accuracy_ratio(obs_bins, fct_bins)


@assert_length
@drop_nan
def calc_threshold_miss_ratio(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    threshold: Number,
    compare: str = ">=",
) -> float:
    """Calculate missing ratio of forecast filtered by threshold.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.
        threshold (Number): The threshold to filter obervation and forecast.
                            This parameter should be used with `compare` parameter,
                            The `compare` parameter will control the logical operator.
        compare (str, optional): The logical operator applying `threshold` parameter.
                                * '>' greater
                                * '<' less
                                * '>=' greater or equal
                                * '<=' less or equal
                                Defaults to '>='.

    Returns:
        float: The missing ratio of forecast filtered by threshold.
    """
    obs_bins, fct_bins = threshold_binarize(
        observation, forecast, threshold=threshold, compare=compare
    )

    return calc_miss_ratio(obs_bins, fct_bins)


@assert_length
@drop_nan
def calc_threshold_false_alarm_ratio(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    threshold: Number,
    compare: str = ">=",
) -> float:
    """Calculate false alarm ratio of forecast filtered by threshold.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.
        threshold (Number): The threshold to filter obervation and forecast.
                            This parameter should be used with `compare` parameter,
                            The `compare` parameter will control the logical operator.
        compare (str, optional): The logical operator applying `threshold` parameter.
                                * '>' greater
                                * '<' less
                                * '>=' greater or equal
                                * '<=' less or equal
                                Defaults to '>='.

    Returns:
        float: The false alarm ratio of forecast filtered by threshold.
    """
    obs_bins, fct_bins = threshold_binarize(
        observation, forecast, threshold=threshold, compare=compare
    )

    return calc_false_alarm_ratio(obs_bins, fct_bins)


@assert_length
@drop_nan
def calc_threshold_bias_score(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    threshold: Number,
    compare: str = ">=",
) -> float:
    """Calculate bias score of forecast filtered by threshold.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.
        threshold (Number): The threshold to filter obervation and forecast.
                            This parameter should be used with `compare` parameter,
                            The `compare` parameter will control the logical operator.
        compare (str, optional): The logical operator applying `threshold` parameter.
                                * '>' greater
                                * '<' less
                                * '>=' greater or equal
                                * '<=' less or equal
                                Defaults to '>='.

    Returns:
        float: The bias score of forecast filtered by threshold.
    """
    obs_bins, fct_bins = threshold_binarize(
        observation, forecast, threshold=threshold, compare=compare
    )

    return calc_bias_score(obs_bins, fct_bins)


@assert_length
@drop_nan
def calc_threshold_ts(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    threshold: Number,
    compare: str = ">=",
) -> float:
    """Calculate Threat score(TS) of forecast filtered by threshold.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.
        threshold (Number): The threshold to filter obervation and forecast.
                            This parameter should be used with `compare` parameter,
                            The `compare` parameter will control the logical operator.
        compare (str, optional): The logical operator applying `threshold` parameter.
                                * '>' greater
                                * '<' less
                                * '>=' greater or equal
                                * '<=' less or equal
                                Defaults to '>='.

    Returns:
        float: The Threat score(TS) of forecast filtered by threshold.
    """
    obs_bins, fct_bins = threshold_binarize(
        observation, forecast, threshold=threshold, compare=compare
    )

    return calc_ts(obs_bins, fct_bins)


@assert_length
@drop_nan
def calc_threshold_mae(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    threshold: Number,
    compare: str = ">=",
) -> float:
    """Calculate MAE(mean absolute error) of forecast filtered by threshold.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.
        threshold (Number): The threshold to filter obervation and forecast.
                            This parameter should be used with `compare` parameter,
                            The `compare` parameter will control the logical operator.
        compare (str, optional): The logical operator applying `threshold` parameter.
                                * '>' greater
                                * '<' less
                                * '>=' greater or equal
                                * '<=' less or equal
                                Defaults to '>='.

    Returns:
        float: The MAE(mean absolute error) of forecast filtered by threshold.
    """
    obs_bins = np.full_like(observation, False)
    fct_bins = np.full_like(forecast, False)

    exec(f"obs_index = observation {compare} {threshold}")
    exec(f"fct_index = forecast {compare} {threshold}")

    obs_bins[locals()["obs_index"]] = True
    fct_bins[locals()["fct_index"]] = True

    return calc_mae(obs_bins, fct_bins)



import numpy as np

from .wind_directions import (
    LEAST_ANGLE_DEFLECTION_CASE,
    LEAST_DIR_DEFLECTION_CASE,
    IDENTIFY_DIRECTION8_CASE,
    WIND_DIR_SCORE_CASE,
    WIND_DIR_ACCURACY_RATE,
)
from .wind_speed import (
    IDENTIFY_SPEED_LEVEL_GENERAL_CASE,
    IDENTIFY_SPEED_LEVEL_MAPPING_CASE,
    LEAST_LEV_DEFLECTION_CASE,
    WIND_LEVEL_ACCURACY_RATE_CASE,
    WIND_SPEED_SCORE_CASE,
    WIND_SPEED_ACCURACY_RATE_CASE,
    WIND_SPEED_MAE_CASE,
    WIND_SPEED_STRONGER_RATE_CASE,
    WIND_SPEED_WEAKER_RATE_CASE,
    WIND_SPEED_CHI_SQUARE_CASE,
    WIND_SPEED_RESIDUAL_SUM_OF_SQUARE_CASE,
    WIND_SPEED_LINREGRESS_CASE,
)

FILTER_WIND_SCALES_CASE = [
    {
        "obs": np.array(
            [
                5.96475312,
                2.9787626,
                3.8667877,
                14.75895423,
                8.48229788,
                10.05979194,
                0.47507922,
                6.50696602,
                16.67486185,
                14.4919493,
            ]
        ),
        "fct": np.array(
            [
                16.238759,
                10.40696757,
                8.82104982,
                3.61639317,
                5.3269034,
                10.9892586,
                7.59016338,
                17.7476437,
                15.76840345,
                1.84942086,
            ]
        ),
        "arg_cases": [
            {
                "parameters": {"scale_min": 3, "scale_max": 5, "return_index": False},
                "result": (np.array([3.9, 8.5]), np.array([8.8, 5.3])),
            },
            {
                "parameters": {"scale_min": 3, "scale_max": 5, "return_index": True},
                "result": (np.array([2, 4]),),
            },
            {
                "parameters": {"scale_min": 4, "scale_max": 10, "return_index": False},
                "result": (
                    np.array([6.0, 10.1, 6.5, 16.7]),
                    np.array([16.2, 11.0, 17.7, 15.8]),
                ),
            },
            {
                "parameters": {"scale_min": 4, "scale_max": 10, "return_index": True},
                "result": (np.array([0, 5, 7, 8]),),
            },
            {
                "parameters": {
                    "scale_min": 5,
                    "scale_max": np.inf,
                    "return_index": False,
                },
                "result": (np.array([10.1, 16.7]), np.array([11.0, 15.8])),
            },
            {
                "parameters": {
                    "scale_min": 5,
                    "scale_max": np.inf,
                    "return_index": True,
                },
                "result": (np.array([5, 8]),),
            },
            {
                "parameters": {
                    "scale_min": 0,
                    "scale_max": 4,
                    "return_index": False,
                },
                "result": (np.array([0.5]), np.array([7.6])),
            },
            {
                "parameters": {
                    "scale_min": 0,
                    "scale_max": 4,
                    "return_index": True,
                },
                "result": (np.array([6]),),
            },
            {
                "parameters": {
                    "scale_min": 0,
                    "scale_max": 4,
                    "return_index": False,
                    "mode": "or",
                },
                "result": (
                    np.array([6.0, 3.0, 3.9, 14.8, 8.5, 0.5, 6.5, 14.5]),
                    np.array([16.2, 10.4, 8.8, 3.6, 5.3, 7.6, 17.7, 1.8]),
                ),
            },
            {
                "parameters": {
                    "scale_min": 0,
                    "scale_max": 4,
                    "return_index": True,
                    "mode": "or",
                },
                "result": (np.array([0, 1, 2, 3, 4, 6, 7, 9]),),
            },
            {
                "parameters": {
                    "scale_min": 8,
                    "return_index": False,
                    "mode": "or",
                },
                "result": (np.array([6.5]), np.array([17.7])),
            },
            {
                "parameters": {
                    "scale_min": 8,
                    "return_index": True,
                    "mode": "or",
                },
                "result": (np.array([7]),),
            },
            {
                "parameters": {
                    "scale_min": 6,
                    "mode": "or",
                },
                "result": (
                    np.array([6.0, 14.8, 10.1, 6.5, 16.7, 14.5]),
                    np.array([16.2, 3.6, 11.0, 17.7, 15.8, 1.8]),
                ),
            },
            {
                "parameters": {
                    "scale_min": 6,
                    "mode": "and",
                },
                "result": (np.array([16.7]), np.array([15.8])),
            },
        ],
    },
]



from cyeva.core.wind import (
    get_least_angle_deflection,
    identify_direction,
    identify_wind_scale,
    get_least_lev_diff,
    get_least_dir_deflection,
    filter_wind_scales,
    WindComparison,
)

from .case.wind import (
    LEAST_ANGLE_DEFLECTION_CASE,
    LEAST_DIR_DEFLECTION_CASE,
    IDENTIFY_DIRECTION8_CASE,
    WIND_DIR_SCORE_CASE,
    WIND_DIR_ACCURACY_RATE,
    IDENTIFY_SPEED_LEVEL_GENERAL_CASE,
    LEAST_LEV_DEFLECTION_CASE,
    WIND_LEVEL_ACCURACY_RATE_CASE,
    WIND_SPEED_SCORE_CASE,
    WIND_SPEED_ACCURACY_RATE_CASE,
    WIND_SPEED_MAE_CASE,
    WIND_SPEED_STRONGER_RATE_CASE,
    WIND_SPEED_WEAKER_RATE_CASE,
    WIND_SPEED_CHI_SQUARE_CASE,
    WIND_SPEED_RESIDUAL_SUM_OF_SQUARE_CASE,
    WIND_SPEED_LINREGRESS_CASE,
    FILTER_WIND_SCALES_CASE,
)


def test_get_least_angle_deflection():
    for case in LEAST_ANGLE_DEFLECTION_CASE:
        angle1 = case["angle1"]
        angle2 = case["angle2"]
        result = case["result"]
        assert get_least_angle_deflection(angle1, angle2) == result


def test_get_least_dir_deflection():
    for case in LEAST_DIR_DEFLECTION_CASE:
        dir1 = case["dir1"]
        dir2 = case["dir2"]
        result = case["result"]
        assert get_least_dir_deflection(dir1, dir2) == result


def test_identify_direction8():
    for case in IDENTIFY_DIRECTION8_CASE:
        angle = case["angle"]
        result = case["result"]
        assert identify_direction(angle, dnum=8) == result


def test_identify_wind_scale():
    for case in IDENTIFY_SPEED_LEVEL_GENERAL_CASE:
        wspd = case["wspd"]
        result = case["result"]
        assert identify_wind_scale(wspd) == result


def test_calc_wind_dir_score():
    for case in WIND_DIR_SCORE_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]

        wind_comparison = WindComparison(obs_dir=obs, fct_dir=fct)

        assert wind_comparison.calc_dir_score() == result


def test_get_least_lev_deflection2():
    for case in LEAST_LEV_DEFLECTION_CASE:
        lev1 = case["lev1"]
        lev2 = case["lev2"]
        result = case["result"]
        assert get_least_lev_diff(lev1, lev2) == result


def test_calc_wind_dir_accuracy_ratio():
    for case in WIND_DIR_ACCURACY_RATE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]

        wind_comparison = WindComparison(obs_dir=obs, fct_dir=fct)

        assert wind_comparison.calc_dir_accuracy_ratio() == result


def test_calc_wind_scale_accuracy_ratio():
    for case in WIND_LEVEL_ACCURACY_RATE_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]

        wind_comparison = WindComparison(obs_spd=obs, fct_spd=fct)

        assert wind_comparison.calc_wind_scale_accuracy_ratio() == result


def test_calc_wind_speed_accuracy_ratio():
    for case in WIND_SPEED_ACCURACY_RATE_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]

        wind_comparison = WindComparison(obs_spd=obs, fct_spd=fct)

        assert wind_comparison.calc_speed_accuracy_ratio() == result


def test_calc_wind_speed_score():
    for case in WIND_SPEED_SCORE_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]

        wind_comparison = WindComparison(obs_spd=obs, fct_spd=fct)

        assert wind_comparison.calc_speed_score() == result


def test_calc_wind_scale_stronger_ratio():
    for case in WIND_SPEED_STRONGER_RATE_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]

        wind_comparison = WindComparison(obs_spd=obs, fct_spd=fct)

        assert wind_comparison.calc_wind_scale_stronger_ratio() == result


def test_calc_wind_scale_weaker_ratio():
    for case in WIND_SPEED_WEAKER_RATE_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]

        wind_comparison = WindComparison(obs_spd=obs, fct_spd=fct)

        assert wind_comparison.calc_wind_scale_weaker_ratio() == result


def test_calc_wind_speed_chi_square():
    for case in WIND_SPEED_CHI_SQUARE_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]

        wind_comparison = WindComparison(obs_spd=obs, fct_spd=fct)

        assert wind_comparison.calc_chi_square() == result


def test_calc_wind_speed_rss():
    for case in WIND_SPEED_RESIDUAL_SUM_OF_SQUARE_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]

        wind_comparison = WindComparison(obs_spd=obs, fct_spd=fct)

        assert wind_comparison.calc_rss() == result


def test_calc_wind_speed_mae():
    for case in WIND_SPEED_MAE_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]

        wind_comparison = WindComparison(obs_spd=obs, fct_spd=fct)

        assert wind_comparison.calc_mae() == result


def test_calc_wind_speed_linregress():
    for case in WIND_SPEED_LINREGRESS_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]

        wind_comparison = WindComparison(obs_spd=obs, fct_spd=fct)

        slope, intercept, *_ = wind_comparison.calc_linregress_args()

        assert slope == result["slope"]
        assert intercept == result["intercept"]


def test_filter_wind_scales():
    for base_case in FILTER_WIND_SCALES_CASE:
        obs = base_case["obs"]
        fct = base_case["fct"]

        for arg_case in base_case["arg_cases"]:
            parameters = arg_case["parameters"]
            result = arg_case["result"]
            _result = filter_wind_scales(obs, fct, **parameters)
            for i in range(len(result)):
                assert (_result[i] == result[i]).all()



from .accuracy_ratio import ACCURACY_RATE_CASE
from .rmse import RMSE_CASE
from .mae import MAE_CASE
from .rss import RESIDUAL_SUM_OF_SQUARE_CASE
from .chi_square import CHI_SQUARE_CASE
from .linregress import LINREGRESS_CASE



from functools import wraps
from numbers import Number

import numpy as np


def assert_length(func):
    """Check if array is empty"""

    @wraps(func)
    def wrapper(observation, forecast, *args, **kwargs):
        if isinstance(observation, Number) and isinstance(forecast, Number):
            return func(observation, forecast, *args, **kwargs)
        if (
            hasattr(observation, "__len__")
            and hasattr(forecast, "__len__")
            and len(observation) == len(forecast)
            and len(observation) > 0
        ):
            return func(observation, forecast, *args, **kwargs)
        else:
            return None

    return wrapper


def convert_to_ndarray(func):
    """Convert list array to numpy ndarray"""

    @wraps(func)
    def wrapper(observation, forecast, *args, **kwargs):

        if not isinstance(observation, np.ndarray) and not isinstance(
            observation, Number
        ):
            observation = np.array(observation)

        if not isinstance(forecast, np.ndarray) and not isinstance(forecast, Number):
            forecast = np.array(forecast)

        return func(observation, forecast, *args, **kwargs)

    return wrapper


def fix_zero_division(func):
    """Check if there is zero division and fix it."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ZeroDivisionError:
            return np.nan

    return wrapper


def drop_nan(func):
    """Drop nan"""

    @wraps(func)
    def wrapper(observation, forecast, *args, **kwargs):
        where_obs_nan = observation != observation
        where_fct_nan = forecast != forecast

        either_nan = where_obs_nan | where_fct_nan

        valid_obs = observation[~either_nan]
        valid_fct = forecast[~either_nan]

        return func(valid_obs, valid_fct, *args, **kwargs)

    return wrapper


def source_round_digit(series_num=2, digit_num=1):
    """A decorator to round source array's significant digit

    Args:
        digit_num (int, optional): The digit num to round.
                                   Defaults to 2.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            args_list = list(args)
            serieses = args[:series_num]
            for i, series in enumerate(serieses):
                try:
                    args_list[i] = np.round(series, digit_num)
                except TypeError:
                    args_list[i] = series
            return func(*args_list, **kwargs)

        return wrapper

    return decorator


def result_round_digit(digit_num=2):
    """A decorator to round result array's significant digit

    Args:
        digit_num (int, optional): The digit num to round.
                                   Defaults to 2.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            if result:
                return round(result, digit_num)
            else:
                return result

        return wrapper

    return decorator



import numpy as np

from cyeva import (
    calc_precip_accumulate_indicators,
    calc_precip_occur_indicators,
    calc_precip_interval_indicators,
    PrecipitationComparison,
)

from .case.precip.single import (
    ACCURACY_RATE_CASE,
    HIT_RATIO_CASE,
    BIAS_SCORE_CASE,
    ETS_SCORE_CASE,
    TS_SCORE_CASE,
    FALSE_ALARM_RATE_CASE,
    FALSE_ALARM_RATIO_CASE,
    MISS_RATE_CASE,
)
from .case.precip.interval import (
    LEV_ACCURACY_RATE_CASE,
    LEV_HIT_RATIO_CASE,
    LEV_TS_SCORE_CASE,
    LEV_ETS_SCORE_CASE,
    LEV_MISS_RATE_CASE,
    LEV_FALSE_ALARM_RATIO_CASE,
    LEV_FALSE_ALARM_RATE_CASE,
    LEV_BIAS_SCORE_CASE,
)
from .case.precip.accumulate import (
    ACC_ACCURACY_RATE_CASE,
    ACC_HIT_RATIO_CASE,
    ACC_TS_SCORE_CASE,
    ACC_ETS_SCORE_CASE,
    ACC_MISS_RATIO_CASE,
    ACC_FALSE_ALARM_RATE_CASE,
    ACC_FALSE_ALARM_RATIO_CASE,
    ACC_BIAS_SCORE_CASE,
)
from .case import (
    RMSE_CASE,
    CHI_SQUARE_CASE,
    MAE_CASE,
    RESIDUAL_SUM_OF_SQUARE_CASE,
    LINREGRESS_ARGS_CASE,
    BIAS_CASE,
    DIFF_ACCURACY_RATIO,
)


def test_calc_accuracy_ratio():
    for case in ACCURACY_RATE_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]
        _result = calc_precip_occur_indicators(obs, fct, indicator="accuracy_ratio")
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)


def test_calc_miss_ratio():
    for case in HIT_RATIO_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]
        _result = calc_precip_occur_indicators(obs, fct, indicator="hit_ratio")
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)


def test_calc_miss_ratio():
    for case in MISS_RATE_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]
        _result = calc_precip_occur_indicators(obs, fct, indicator="miss_ratio")
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)


def test_calc_false_alarm_ratio():
    for case in FALSE_ALARM_RATIO_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]
        _result = calc_precip_occur_indicators(obs, fct, indicator="false_alarm_ratio")
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)


def test_calc_false_alarm_rate():
    for case in FALSE_ALARM_RATE_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]
        _result = calc_precip_occur_indicators(obs, fct, indicator="false_alarm_rate")
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)


def test_calc_ts():
    for case in TS_SCORE_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]
        _result = calc_precip_occur_indicators(obs, fct, indicator="ts")
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)


def test_calc_ets():
    for case in ETS_SCORE_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]
        _result = calc_precip_occur_indicators(obs, fct, indicator="ets")
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)


def test_calc_bias_score():
    for case in BIAS_SCORE_CASE:
        obs = case["obs"]
        fct = case["fct"]
        result = case["result"]
        _result = calc_precip_occur_indicators(obs, fct, indicator="bias_score")
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)


def test_calc_precip_interval_accuracy_ratio():
    for kind, levs in LEV_ACCURACY_RATE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fct = case["fct"]
                result = case["result"]
                _result = calc_precip_interval_indicators(
                    obs, fct, kind, lev, indicator="accuracy_ratio"
                )
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)


def test_calc_precip_interval_miss_ratio():
    for kind, levs in LEV_MISS_RATE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fct = case["fct"]
                result = case["result"]
                _result = calc_precip_interval_indicators(
                    obs, fct, kind, lev, indicator="miss_ratio"
                )
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)


def test_calc_precip_interval_false_alarm_ratio():
    for kind, levs in LEV_FALSE_ALARM_RATIO_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fct = case["fct"]
                result = case["result"]
                _result = calc_precip_interval_indicators(
                    obs, fct, kind, lev, indicator="false_alarm_ratio"
                )
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)


def test_calc_precip_interval_bias_score():
    for kind, levs in LEV_BIAS_SCORE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fct = case["fct"]
                result = case["result"]
                _result = calc_precip_interval_indicators(
                    obs, fct, kind, lev, indicator="bias_score"
                )
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)


def test_calc_precip_interval_hit_ratio():
    for kind, levs in LEV_HIT_RATIO_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fct = case["fct"]
                result = case["result"]
                _result = calc_precip_interval_indicators(
                    obs, fct, kind, lev, indicator="hit_ratio"
                )
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)


def test_calc_precip_interval_ts():
    for kind, levs in LEV_TS_SCORE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fct = case["fct"]
                result = case["result"]
                _result = calc_precip_interval_indicators(
                    obs, fct, kind, lev, indicator="ts"
                )
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)


def test_calc_precip_interval_ets():
    for kind, levs in LEV_ETS_SCORE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fct = case["fct"]
                result = case["result"]
                _result = calc_precip_interval_indicators(
                    obs, fct, kind, lev, indicator="ets"
                )
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)


def test_calc_precip_accumulate_accuracy_ratio():
    for kind, levs in ACC_ACCURACY_RATE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fct = case["fct"]
                result = case["result"]
                _result = calc_precip_accumulate_indicators(
                    obs, fct, kind, lev, indicator="accuracy_ratio"
                )
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)

def test_calc_precip_accumulate_hit_ratio():
    for kind, levs in ACC_HIT_RATIO_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fct = case["fct"]
                result = case["result"]
                _result = calc_precip_accumulate_indicators(
                    obs, fct, kind, lev, indicator="hit_ratio"
                )
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)


def test_calc_precip_accumulate_miss_ratio():
    for kind, levs in ACC_MISS_RATIO_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fct = case["fct"]
                result = case["result"]
                _result = calc_precip_accumulate_indicators(
                    obs, fct, kind, lev, indicator="miss_ratio"
                )
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)


def test_calc_precip_accumulate_false_alarm_ratio():
    for kind, levs in ACC_FALSE_ALARM_RATIO_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fct = case["fct"]
                result = case["result"]
                _result = calc_precip_accumulate_indicators(
                    obs, fct, kind, lev, indicator="false_alarm_ratio"
                )
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)


def test_calc_precip_accumulate_false_alarm_rate():
    for kind, levs in ACC_FALSE_ALARM_RATE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fct = case["fct"]
                result = case["result"]
                _result = calc_precip_accumulate_indicators(
                    obs, fct, kind, lev, indicator="false_alarm_rate"
                )
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)


def test_calc_precip_accumulate_bias_score():
    for kind, levs in ACC_BIAS_SCORE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fct = case["fct"]
                result = case["result"]
                _result = calc_precip_accumulate_indicators(
                    obs, fct, kind, lev, indicator="bias_score"
                )
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)


def test_calc_precip_accumulate_ts():
    for kind, levs in ACC_TS_SCORE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fct = case["fct"]
                result = case["result"]
                _result = calc_precip_accumulate_indicators(
                    obs, fct, kind, lev, indicator="ts"
                )
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)


def test_calc_precip_accumulate_ets():
    for kind, levs in ACC_ETS_SCORE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fct = case["fct"]
                result = case["result"]
                _result = calc_precip_accumulate_indicators(
                    obs, fct, kind, lev, indicator="ets"
                )
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)


def test_precipitation_object_each():
    for case in RMSE_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        precip = PrecipitationComparison(obs, fcst)

        _result = precip.calc_rmse()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in MAE_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        precip = PrecipitationComparison(obs, fcst)

        _result = precip.calc_mae()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in CHI_SQUARE_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        precip = PrecipitationComparison(obs, fcst)

        _result = precip.calc_chi_square()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in RESIDUAL_SUM_OF_SQUARE_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        precip = PrecipitationComparison(obs, fcst)

        _result = precip.calc_rss()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in LINREGRESS_ARGS_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        precip = PrecipitationComparison(obs, fcst)

        _result = precip.calc_linregress_args()
        assert _result == result

    for case in BIAS_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        precip = PrecipitationComparison(obs, fcst)

        _result = precip.calc_bias()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in DIFF_ACCURACY_RATIO:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        precip = PrecipitationComparison(obs, fcst)

        _result = precip.calc_diff_accuracy_ratio()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in ACCURACY_RATE_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        precip = PrecipitationComparison(obs, fcst)

        _result = precip.calc_accuracy_ratio()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in ACCURACY_RATE_CASE:
        obs = np.array(case["obs"]) / 1000
        fcst = np.array(case["fct"]) / 1000
        result = case["result"]

        precip = PrecipitationComparison(obs, fcst, unit="m")

        _result = precip.calc_accuracy_ratio()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in MISS_RATE_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        precip = PrecipitationComparison(obs, fcst)

        _result = precip.calc_miss_ratio()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in FALSE_ALARM_RATIO_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        precip = PrecipitationComparison(obs, fcst)

        _result = precip.calc_false_alarm_ratio()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in TS_SCORE_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        precip = PrecipitationComparison(obs, fcst)

        _result = precip.calc_ts()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in ETS_SCORE_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        precip = PrecipitationComparison(obs, fcst)

        _result = precip.calc_ets()

        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in BIAS_SCORE_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        precip = PrecipitationComparison(obs, fcst)

        _result = precip.calc_bias_score()

        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for kind, levs in LEV_ACCURACY_RATE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fcst = case["fct"]
                result = case["result"]

                precip = PrecipitationComparison(obs, fcst)

                _result = precip.calc_accuracy_ratio(lev=str(lev), kind=kind)

                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)

    for kind, levs in LEV_MISS_RATE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fcst = case["fct"]
                result = case["result"]

                precip = PrecipitationComparison(obs, fcst)

                _result = precip.calc_miss_ratio(lev=str(lev), kind=kind)

                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)

    for kind, levs in LEV_FALSE_ALARM_RATIO_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fcst = case["fct"]
                result = case["result"]
                precip = PrecipitationComparison(obs, fcst)

                _result = precip.calc_false_alarm_ratio(lev=str(lev), kind=kind)

                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)

    for kind, levs in LEV_BIAS_SCORE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fcst = case["fct"]
                result = case["result"]
                precip = PrecipitationComparison(obs, fcst)

                _result = precip.calc_bias_score(lev=str(lev), kind=kind)

                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)

    for kind, levs in LEV_TS_SCORE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fcst = case["fct"]
                result = case["result"]
                precip = PrecipitationComparison(obs, fcst)

                _result = precip.calc_ts(lev=str(lev), kind=kind)
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)

    for kind, levs in LEV_ETS_SCORE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fcst = case["fct"]
                result = case["result"]
                precip = PrecipitationComparison(obs, fcst)

                _result = precip.calc_ets(lev=str(lev), kind=kind)
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)

    for kind, levs in ACC_ACCURACY_RATE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fcst = case["fct"]
                result = case["result"]

                precip = PrecipitationComparison(obs, fcst)

                _result = precip.calc_accuracy_ratio(lev="+" + str(lev), kind=kind)

                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)

    for kind, levs in ACC_MISS_RATIO_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fcst = case["fct"]
                result = case["result"]

                precip = PrecipitationComparison(obs, fcst)

                _result = precip.calc_miss_ratio(lev="+" + str(lev), kind=kind)

                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)

    for kind, levs in ACC_FALSE_ALARM_RATIO_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fcst = case["fct"]
                result = case["result"]
                precip = PrecipitationComparison(obs, fcst)

                _result = precip.calc_false_alarm_ratio(lev="+" + str(lev), kind=kind)

                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)

    for kind, levs in ACC_FALSE_ALARM_RATE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fcst = case["fct"]
                result = case["result"]
                precip = PrecipitationComparison(obs, fcst)

                _result = precip.calc_false_alarm_rate(lev="+" + str(lev), kind=kind)

                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)

    for kind, levs in ACC_BIAS_SCORE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fcst = case["fct"]
                result = case["result"]
                precip = PrecipitationComparison(obs, fcst)

                _result = precip.calc_bias_score(lev="+" + str(lev), kind=kind)

                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)

    for kind, levs in ACC_TS_SCORE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fcst = case["fct"]
                result = case["result"]
                precip = PrecipitationComparison(obs, fcst)

                _result = precip.calc_ts(lev="+" + str(lev), kind=kind)
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)

    for kind, levs in ACC_ETS_SCORE_CASE.items():
        for lev, cases in levs.items():
            for case in cases:
                obs = case["obs"]
                fcst = case["fct"]
                result = case["result"]
                precip = PrecipitationComparison(obs, fcst)

                _result = precip.calc_ets(lev="+" + str(lev), kind=kind)
                if not np.isnan(result):
                    assert _result == result
                else:
                    assert np.isnan(_result)


def test_precipitation_object_all():
    for case in ACCURACY_RATE_CASE:
        obs = case["obs"]
        fcst = case["fct"]

        precip = PrecipitationComparison(obs, fcst)

        precip.gather_all_indicators(kind="1h")
        precip.gather_all_indicators(kind="3h")
        precip.gather_all_indicators(kind="24h")


if __name__ == "__main__":
    test_precipitation_object_all()



import warnings
from pint import UnitStrippedWarning

warnings.filterwarnings("ignore", category=UnitStrippedWarning)

from .base import *
from .precip import *
from .temp import *
from .wind import *



from typing import Union
from numbers import Number

import numpy as np

from ..utils.decorators import (
    convert_to_ndarray,
    assert_length,
    drop_nan,
    source_round_digit,
)


@source_round_digit(digit_num=2)
@convert_to_ndarray
@assert_length
@drop_nan
def threshold_binarize(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    threshold: Number = 0,
    compare: str = ">=",
) -> tuple:
    """Binarize observation and forecast array filtered by threshold.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.

        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

        threshold (Number): The threshold to filter obervation and forecast.
                            This parameter should be used with `compare` parameter,
                            The `compare` parameter will control the logical operator.

        compare (str, optional): The logical operator applying to `threshold` parameter.
                                * '>' greater
                                * '<' less
                                * '>=' greater or equal
                                * '<=' less or equal
                                Defaults to '>='.

    Returns:
        tuple: Binarized array tuple of (observation, forecast).
    """

    obs_bins = np.full_like(observation, False)
    fct_bins = np.full_like(forecast, False)

    exec(f"obs_index = observation {compare} {threshold}")
    exec(f"fct_index = forecast {compare} {threshold}")

    obs_bins[locals()["obs_index"]] = True
    fct_bins[locals()["fct_index"]] = True

    obs_bins = obs_bins.astype(bool)
    fct_bins = fct_bins.astype(bool)

    return obs_bins, fct_bins


@source_round_digit(digit_num=2)
@convert_to_ndarray
@assert_length
@drop_nan
def level_binarize(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    lev_config: dict,
    lev_id: int,
    include_zero=False,
) -> tuple:
    """Binarize observation and forecast array by level.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.

        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

        lev_config (dict): Level configurature dict,
                           you can find some samples from cyeva.config.levels,
                           the dict's structure like:

                            lev_config = {
                                1: {"min": 0, "max": 1000},
                                2: {"min": 1000, "max": 5000},
                                3: {"min": 5000, "max": 10000},
                                4: {"min": 10000, "max": np.inf},
                            }

                            lev_config's keys should be the level's id
                            and values should contain max and min values.

        lev_id (int): The selected level id to binarize,
                        namely, this id will be marked as True
                        and other level id will be marked as False.
                        `lev_id` must be in `lev_config`'s keys.

        include_zero (bool): A flag to indicate weather include zero when filtering.
                             Default to False.

    Returns:
        tuple: Binarized array tuple of (observation, forecast) according to level id.
    """
    lev_obs = np.full((len(observation),), np.nan)
    lev_fct = np.full((len(forecast),), np.nan)

    interval = lev_config[lev_id]

    if not set(["max", "min"]).issubset(set(interval.keys())):
        raise ValueError(
            "lev_config's each interval item " "should contains 'max' and 'min' keys."
        )

    minimum = interval["min"]
    maximum = interval["max"]

    if include_zero:
        index_obs = np.where((observation >= minimum) & (observation <= maximum))
        index_fct = np.where((forecast >= minimum) & (forecast <= maximum))
    else:
        if minimum > 0:
            index_obs = np.where((observation >= minimum) & (observation <= maximum))
            index_fct = np.where((forecast >= minimum) & (forecast <= maximum))
        else:
            index_obs = np.where((observation > minimum) & (observation <= maximum))
            index_fct = np.where((forecast > minimum) & (forecast <= maximum))

    lev_obs[index_obs] = lev_id
    lev_fct[index_fct] = lev_id

    lev_obs[~np.isclose(lev_obs, float(lev_id))] = 0
    lev_fct[~np.isclose(lev_fct, float(lev_id))] = 0

    lev_obs = lev_obs.astype(bool)
    lev_fct = lev_fct.astype(bool)

    return lev_obs, lev_fct



import numpy as np

from cyeva.core.base import Comparison
from .case import (
    RMSE_CASE,
    MAE_CASE,
    CHI_SQUARE_CASE,
    RESIDUAL_SUM_OF_SQUARE_CASE,
    LINREGRESS_ARGS_CASE,
    BINARY_ACCURACY_RATIO_CASE,
    BIAS_CASE,
    DIFF_ACCURACY_RATIO,
)


def test_comparison():

    for case in RMSE_CASE:
        obs = case["obs"]
        fcst = case["fct"]

        cp = Comparison(obs, fcst)

        result = case["result"]
        _result = cp.calc_rmse()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in MAE_CASE:
        obs = case["obs"]
        fcst = case["fct"]

        cp = Comparison(obs, fcst)

        result = case["result"]
        _result = cp.calc_mae()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in CHI_SQUARE_CASE:
        obs = case["obs"]
        fcst = case["fct"]

        cp = Comparison(obs, fcst)

        result = case["result"]
        _result = cp.calc_chi_square()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in RESIDUAL_SUM_OF_SQUARE_CASE:
        obs = case["obs"]
        fcst = case["fct"]

        cp = Comparison(obs, fcst)

        result = case["result"]
        _result = cp.calc_rss()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in LINREGRESS_ARGS_CASE:
        obs = case["obs"]
        fcst = case["fct"]

        cp = Comparison(obs, fcst)

        result = case["result"]
        _result = cp.calc_linregress_args()
        assert _result == result

    for case in BINARY_ACCURACY_RATIO_CASE:
        obs = case["obs"]
        fcst = case["fct"]

        cp = Comparison(obs, fcst)

        result = case["result"]
        _result = cp.calc_binary_accuracy_ratio()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in BIAS_CASE:
        obs = case["obs"]
        fcst = case["fct"]

        cp = Comparison(obs, fcst)

        result = case["result"]
        _result = cp.calc_bias()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)

    for case in DIFF_ACCURACY_RATIO:
        obs = case["obs"]
        fcst = case["fct"]

        cp = Comparison(obs, fcst)

        result = case["result"]
        _result = cp.calc_diff_accuracy_ratio()
        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)



from .lev_accuracy_ratio import LEV_ACCURACY_RATE_CASE
from .lev_hit_ratio import LEV_HIT_RATIO_CASE
from .lev_ts import LEV_TS_SCORE_CASE
from .lev_miss_ratio import LEV_MISS_RATE_CASE
from .lev_ets import LEV_ETS_SCORE_CASE
from .lev_bias_score import LEV_BIAS_SCORE_CASE
from .lev_false_alarm import LEV_FALSE_ALARM_RATIO_CASE, LEV_FALSE_ALARM_RATE_CASE



from typing import List, Union
from numbers import Number
from collections import Counter
from functools import partial

import numpy as np
import pandas as pd
from pint import UnitRegistry, Quantity

from ..utils import (
    convert_to_ndarray,
    fix_zero_division,
    result_round_digit,
    source_round_digit,
    assert_length,
    drop_nan,
)
from ..config.directions.wind import (
    DIRECTION8_CENTER_ANGLE,
    DIRECTION16_CENTER_ANGLE,
)
from ..config.levels.wind import (
    GENERAL_WIND_SCALES,
)

from ..core import Comparison

UNITS = UnitRegistry()


def get_least_angle_deflection(
    angle1: Union[Number, np.ndarray], angle2: Union[Number, np.ndarray]
) -> Union[Number, np.ndarray]:
    """Calculate the least angle deflection between two angles.

    Args:
        angle1 (Union[Number, np.ndarray]): One angle(degree).
        angle2 (Union[Number, np.ndarray]): Another angle(degree).

    Returns:
        Union[Number, np.ndarray]: The least deflection between two angles.
    """

    if isinstance(angle1, Quantity):
        angle1_magnitude = angle1.magnitude
    else:
        angle1_magnitude = angle1
    if isinstance(angle2, Quantity):
        angle2_magnitude = angle2.magnitude
    else:
        angle2_magnitude = angle2

    angle1_magnitude %= 360
    angle2_magnitude %= 360

    deflection1 = abs(angle1_magnitude - angle2_magnitude)
    deflection2 = 360 - abs(angle1_magnitude - angle2_magnitude)

    if isinstance(angle1, Number) and isinstance(angle2, Number):
        return deflection1 if deflection1 < deflection2 else deflection2
    else:
        index1 = np.where(deflection1 <= deflection2)
        index2 = np.where(deflection2 <= deflection1)

        deflection = np.full_like(deflection1, 0)

        deflection[index1] = deflection1[index1]
        deflection[index2] = deflection2[index2]

        return deflection


def identify_direction(
    angle: Union[Number, np.ndarray], dnum: int = 8
) -> Union[int, np.ndarray]:
    """Identify 8 cardinal directions by angle.

    Args:
        angle (Number): The wind direction in degree.
        dnum (int): The wind directions number.

    Returns:
        Union[int, np.ndarray]: Direction ID of the cardinal direction.
    """
    if isinstance(angle, List):
        angle = np.array(angle)

    if isinstance(angle, np.ndarray) or isinstance(angle, Quantity):
        angle_magnitude = angle.magnitude

    if isinstance(angle, Number):
        angle_magnitude = angle

    if dnum == 8:
        DIRECTION_CENTER_ANGLE = DIRECTION8_CENTER_ANGLE
        threshold = 22.5
    elif dnum == 16:
        DIRECTION_CENTER_ANGLE = DIRECTION16_CENTER_ANGLE
        threshold = 11.25

    angle_magnitude %= 360

    if isinstance(angle, Number):
        for dir_id, center_angle in DIRECTION_CENTER_ANGLE.items():
            if get_least_angle_deflection(angle_magnitude, center_angle) <= threshold:
                break

        return dir_id

    elif isinstance(angle, np.ndarray) or isinstance(angle, Quantity):
        dir_ids = np.full_like(angle_magnitude, -1)
        for dir_id, center_angle in DIRECTION_CENTER_ANGLE.items():
            least_angle_defl = get_least_angle_deflection(angle_magnitude, center_angle)
            dir_ids[least_angle_defl <= threshold] = dir_id

        return dir_ids


def identify_wind_scale(speed: Union[Number, np.ndarray]) -> Union[int, np.ndarray]:
    """Identify wind scale by speed.

    Args:
        speed (Union[Number, np.ndarray]): Wind speed in m/s.

    Returns:
        Union[int, np.ndarray]: Wind scale.
    """
    if isinstance(speed, list):
        speed = np.array(speed)
    if isinstance(speed, np.ndarray) or isinstance(speed, Quantity):
        spd_levs = np.full_like(speed, np.nan)

    try:
        speed_magnitude = speed.magnitude
    except AttributeError:
        speed_magnitude = speed

    for lev, attr in GENERAL_WIND_SCALES.items():
        minimum = attr["min"]
        maximum = attr["max"]
        if isinstance(speed, Number):
            if lev == 0:
                if minimum <= speed <= maximum:
                    return lev
            else:
                if minimum < speed <= maximum:
                    return lev
        elif isinstance(speed, np.ndarray) or isinstance(speed, Quantity):
            if lev == 0:
                spd_levs[
                    (speed_magnitude >= minimum) & (speed_magnitude <= maximum)
                ] = lev
            else:
                spd_levs[
                    (speed_magnitude > minimum) & (speed_magnitude <= maximum)
                ] = lev

    return spd_levs


def get_least_lev_diff(
    lev1: Union[int, np.ndarray],
    lev2: Union[int, np.ndarray],
) -> int:
    """Calculate least level difference.

    Args:
        lev1 (int): One level value.
        lev2 (int): Another level value.

    Returns:
        int: Least level difference.
    """
    return np.abs(lev1 - lev2)


def get_least_dir_deflection(
    dir1: Union[int, np.ndarray], dir2: Union[int, np.ndarray], circle_num: int = 8
) -> Union[int, np.ndarray]:
    """Get least deflection from two directions.

    Args:
        dir1 (Union[int, np.ndarray]): One direction number(0-7 if circle_num is 8)
        dir2 (Union[int, np.ndarray]): Another direction number(0-7 if circle_num is 8)
        circle_num (int, optional): Circulation knot num. Defaults to 8.

    Returns:
        Union[int, np.ndarray]: The least deflection of directions.
    """
    if isinstance(dir1, Number) and isinstance(dir2, Number):
        deflection1 = abs(dir1 - dir2)
        deflection2 = circle_num - abs(dir1 - dir2)

        return int(deflection1) if deflection1 < deflection2 else int(deflection2)
    else:
        deflection1 = np.abs(dir1 - dir2)
        deflection2 = circle_num - np.abs(dir1 - dir2)

        index1 = np.where(deflection1 <= deflection2)
        index2 = np.where(deflection2 <= deflection1)

        deflection = np.full_like(deflection1, 0)

        deflection[index1] = deflection1[index1]
        deflection[index2] = deflection2[index2]

        return deflection


@assert_length
@source_round_digit(series_num=2)
@convert_to_ndarray
@drop_nan
def filter_wind_scales(
    observation,
    forecast,
    scale_min=None,
    scale_max=None,
    mode="and",
    return_index=False,
    *args,
    **kwargs,
):
    """Filter wind by wind scales.

    Args:
        observation (list | ndarray): Observation wind speed array.
        forecast (list | ndarray): Forecast wind speed array.
        scale_min (float | int, optional): The minimum wind scale to filter,
            if set to None it will be the minimum scale in default scales.
            Defaults to None.
        scale_max (float | int, optional): The maximum wind scale to filter,
            if set to None it will be the maximum scale in default scales.
            Defaults to None.
        mode (str, optional): Filter logic(and/or). Defaults to 'and'.
        return_index (boolean, optional): Whether return index, if it is False
            return values, ohterwise return array's index. Defaults to False.

    Returns:
        tuple: (Observation wind speed array, Forecast wind speed array)
    """

    if not scale_min:
        scale_min = float(sorted(GENERAL_WIND_SCALES.keys())[0])
    if not scale_max:
        scale_max = float(sorted(GENERAL_WIND_SCALES.keys())[-1])

    obs_lev = identify_wind_scale(observation)
    fct_lev = identify_wind_scale(forecast)
    assert (obs_lev == obs_lev).all() and (fct_lev == fct_lev).all()

    if mode == "and":
        bools = ((obs_lev >= scale_min) & (obs_lev <= scale_max)) & (
            (fct_lev >= scale_min) & (fct_lev <= scale_max)
        )
        index = np.where(bools == True)
    elif mode == "or":
        bools = ((obs_lev >= scale_min) & (obs_lev <= scale_max)) | (
            (fct_lev >= scale_min) & (fct_lev <= scale_max)
        )
        index = np.where(bools == True)

    filtered_obs = observation[index]
    filtered_fct = forecast[index]

    if return_index:
        return index
    else:
        return filtered_obs, filtered_fct


class WindComparison(Comparison):
    def __init__(
        self,
        obs_spd: Union[np.ndarray, list] = None,
        fct_spd: Union[np.ndarray, list] = None,
        obs_dir: Union[np.ndarray, list] = None,
        fct_dir: Union[np.ndarray, list] = None,
        unit_spd: str = "m/s",
        unit_dir: str = "degree",
    ):
        if obs_spd is not None:
            self.obs_spd = (obs_spd * UNITS.parse_expression(unit_spd)).to("m/s")
        if obs_dir is not None:
            self.obs_dir = (obs_dir * UNITS.parse_expression(unit_dir)).to("degree")
        if fct_spd is not None:
            self.fct_spd = (fct_spd * UNITS.parse_expression(unit_spd)).to("m/s")
        if fct_dir is not None:
            self.fct_dir = (fct_dir * UNITS.parse_expression(unit_dir)).to("degree")

        self.df = pd.DataFrame(
            {
                "observation_speed": obs_spd,
                "forecast_speed": fct_spd,
                "observation_direction": obs_dir,
                "forecast_direction": fct_dir,
            }
        )

    def __repr__(self):
        text = "\n".join(["<Object:WindComparison>", "data:", self.df.__str__()])
        return text

    def __str__(self):
        text = "\n".join(["<Object:WindComparison>", "data:", self.df.__str__()])
        return text

    @result_round_digit(4)
    @source_round_digit()
    @fix_zero_division
    def calc_rmse(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        kind: str = "speed",
        *args,
        **kwargs,
    ):
        "Root Mean Square Error"
        if observation is None:
            if kind == "speed":
                observation = self.obs_spd
            elif kind == "direction":
                observation = self.obs_dir
        if forecast is None:
            if kind == "speed":
                forecast = self.fct_spd
            elif kind == "direction":
                forecast = self.fct_dir

        return super().calc_rmse(observation, forecast)

    @result_round_digit(4)
    @source_round_digit()
    @fix_zero_division
    def calc_mae(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        kind: str = "speed",
        *args,
        **kwargs,
    ):
        """Mean Absolute Error"""
        if observation is None:
            if kind == "speed":
                observation = self.obs_spd
            elif kind == "direction":
                observation = self.obs_dir
        if forecast is None:
            if kind == "speed":
                forecast = self.fct_spd
            elif kind == "direction":
                forecast = self.fct_dir

        return super().calc_mae(observation, forecast)

    @result_round_digit(4)
    @source_round_digit()
    @fix_zero_division
    def calc_chi_square(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        kind: str = "speed",
        *args,
        **kwargs,
    ):
        """χ2"""
        if observation is None:
            if kind == "speed":
                observation = self.obs_spd
            elif kind == "direction":
                observation = self.obs_dir
        if forecast is None:
            if kind == "speed":
                forecast = self.fct_spd
            elif kind == "direction":
                forecast = self.fct_dir

        return super().calc_chi_square(observation, forecast)

    @result_round_digit(4)
    @source_round_digit()
    @fix_zero_division
    def calc_rss(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        kind: str = "speed",
        *args,
        **kwargs,
    ):
        """Residual Sum of Square"""
        if observation is None:
            if kind == "speed":
                observation = self.obs_spd
            elif kind == "direction":
                observation = self.obs_dir
        if forecast is None:
            if kind == "speed":
                forecast = self.fct_spd
            elif kind == "direction":
                forecast = self.fct_dir

        return super().calc_rss(observation, forecast)

    def calc_linregress_args(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        kind: str = "speed",
        *args,
        **kwargs,
    ):
        """linregress args"""
        if observation is None:
            if kind == "speed":
                observation = self.obs_spd
            elif kind == "direction":
                observation = self.obs_dir
        if forecast is None:
            if kind == "speed":
                forecast = self.fct_spd
            elif kind == "direction":
                forecast = self.fct_dir

        return super().calc_linregress_args(observation, forecast)

    @result_round_digit(4)
    @source_round_digit()
    @fix_zero_division
    def calc_bias(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        kind: str = "speed",
        threshold: Number = 0,
        *args,
        **kwargs,
    ):
        """BIAS score"""
        if observation is None:
            if kind == "speed":
                observation = self.obs_spd
            elif kind == "direction":
                observation = self.obs_dir
        if forecast is None:
            if kind == "speed":
                forecast = self.fct_spd
            elif kind == "direction":
                forecast = self.fct_dir

        return super().calc_bias(observation, forecast, threshold=threshold)

    @result_round_digit(4)
    @source_round_digit()
    @fix_zero_division
    def calc_binary_accuracy_ratio(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        kind: str = "speed",
        threshold: Number = 0,
        *args,
        **kwargs,
    ) -> float:
        """Binary accuracy ratio

        Args:
            threshold (Number, optional): The threshold to binarize array.
                                          Defaults to 0.

        Returns:
            float: The accuracy ratio.
        """
        if observation is None:
            if kind == "speed":
                observation = self.obs_spd
            elif kind == "direction":
                observation = self.obs_dir
        if forecast is None:
            if kind == "speed":
                forecast = self.fct_spd
            elif kind == "direction":
                forecast = self.fct_dir

        return super().calc_binary_accuracy_ratio(
            observation, forecast, threshold=threshold
        )

    @result_round_digit(4)
    @source_round_digit()
    @fix_zero_division
    def calc_diff_accuracy_ratio(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        kind: str = "speed",
        limit: Number = 1,
        *args,
        **kwargs,
    ) -> float:
        """
        Calculate the accuracy ratio that
        two array's difference within a limit scope.

        Args:
            limit (Number, optional): The limit of difference.
                                      Defaults to 1.

        Returns:
            float: The accuracy ratio.
        """
        if observation is None:
            if kind == "speed":
                observation = self.obs_spd
            elif kind == "direction":
                observation = self.obs_dir
        if forecast is None:
            if kind == "speed":
                forecast = self.fct_spd
            elif kind == "direction":
                forecast = self.fct_dir

        return super().calc_diff_accuracy_ratio(observation, forecast, limit=limit)

    @result_round_digit(4)
    @source_round_digit()
    @fix_zero_division
    def calc_dir_accuracy_ratio(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        mode="degree",
        threshold=22.5,
        *args,
        **kwargs,
    ) -> float:
        """Calculate wind direction accuracy ratio.

        Args:
            mode (str, optional): Wind direction mode. Options as follow:
                                  * degree: Calculate wind direction accuracy ratio from degree way.
                                  * drange8: Calculate wind direction accuracy ratio from 8 cardinal directions way.
                                  * drange16 Calculate wind direction accuracy ratio from 16 cardinal directions way.
                                  Defaults to "degree".
            threshold (float, optional): The degree threshold to check if wind
                                         is accurate, only useful when
                                         mode=="degree".
                                         Defaults to 22.5.

        Returns:
            float: The accuracy ratio (%).
        """
        if observation is None:
            observation = self.obs_dir
        if forecast is None:
            forecast = self.fct_dir
        if mode == "degree":
            angle_deflection = get_least_angle_deflection(observation, forecast)

            count_series = np.full_like(angle_deflection, 0).astype(float)
            count_series[angle_deflection <= threshold] = 1

            try:
                return np.sum(count_series) / len(count_series) * 100
            except TypeError:
                return np.sum(count_series) * 100

        elif mode == "drange8":
            obs_d8 = identify_direction(observation)
            fct_d8 = identify_direction(forecast)

            cross = obs_d8 == fct_d8
            try:
                total = len(cross)
            except TypeError:
                cross = [cross]
                total = 1
            counter = Counter(cross)
            correct = counter[True]

            return correct / total * 100

        elif mode == "drange16":
            obs_d16 = identify_direction(observation, dnum=16)
            fct_d16 = identify_direction(forecast, dnum=16)

            cross = obs_d16 == fct_d16
            try:
                total = len(cross)
            except TypeError:
                cross = [cross]
                total = 1
            counter = Counter(cross)
            correct = counter[True]

            return correct / total * 100

    @result_round_digit(4)
    @source_round_digit()
    @fix_zero_division
    def calc_dir_score(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        dnum: int = 8,
        *args,
        **kwargs,
    ) -> float:
        """Calculate wind direction score.

        Args:
            observation (Union[Number, np.ndarray]): Observation wind direction value or ndarray in degree.
            forecast (Union[Number, np.ndarray]): Forecast wind direction value or ndarray in degree.
            dnum (int, optional): The wind directions number. Defaults to 8.

        Returns:
            float: The wind direction score
        """
        if observation is None:
            observation = self.obs_dir
        if forecast is None:
            forecast = self.fct_dir

        if dnum == 8:
            obs_d8 = identify_direction(observation)
            fct_d8 = identify_direction(forecast)

            dir_deflection = get_least_lev_diff(obs_d8, fct_d8)

            score_series = np.full_like(dir_deflection, 0).astype(float)
            score_series[np.isclose(dir_deflection, 1)] = 0.6
            score_series[dir_deflection < 1] = 1

        elif dnum == 16:
            obs_d16 = identify_direction(observation, dnum=16)
            fct_d16 = identify_direction(forecast, dnum=16)

            dir_deflection = get_least_lev_diff(obs_d16, fct_d16)

            score_series = np.full_like(dir_deflection, 0).astype(float)
            score_series[np.isclose(dir_deflection, 2)] = 0.6
            score_series[np.isclose(dir_deflection, 1)] = 0.8
            score_series[dir_deflection < 1] = 1

        try:
            return np.sum(score_series) / len(score_series)
        except TypeError:
            return np.sum(score_series)

    @result_round_digit(4)
    @source_round_digit()
    @fix_zero_division
    def calc_wind_scale_accuracy_ratio(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        scale_min=None,
        scale_max=None,
        mode="strict",
    ):
        """Calculate wind scale accuracy ratio

        Args:
            observation (Union[Number, np.ndarray]): Observation wind speed value or ndarray in m/s.
            forecast (Union[Number, np.ndarray]): Forecast wind speed value or ndarray in m/s.
            scale_min (Number): The min wind scale to remain after filter.
            scale_max (Number): The max wind scale to remain after filter.
            mode (str): The mode to filter, Options are 'strict' and 'loose'.
                        If 'strict' it will use 'and' logic when filter observation and forecast.
                        If 'loose' it will use 'or' logic.

        Returns:
            float: Wind scale accuracy ratio (%)
        """
        if observation is None:
            observation = self.obs_spd
        if forecast is None:
            forecast = self.fct_spd

        if mode == "strict":
            observation, forecast = filter_wind_scales(
                observation, forecast, scale_min=scale_min, scale_max=scale_max
            )
        elif mode == "loose":
            observation, forecast = filter_wind_scales(
                observation,
                forecast,
                scale_min=scale_min,
                scale_max=scale_max,
                mode="or",
            )

        obs_lev = identify_wind_scale(observation)
        fct_lev = identify_wind_scale(forecast)

        hits = np.count_nonzero(obs_lev == fct_lev)
        try:
            total = len(observation)
        except TypeError:
            total = 1

        return hits / total * 100

    @result_round_digit(4)
    @source_round_digit()
    @fix_zero_division
    def calc_speed_accuracy_ratio(
        self,
        observation: Union[Number, np.ndarray] = None,
        forecast: Union[Number, np.ndarray] = None,
        limit: Number = 2,
    ):
        if observation is None:
            observation = self.obs_spd
        if forecast is None:
            forecast = self.fct_spd

        return self.calc_diff_accuracy_ratio(
            observation, forecast, kind="speed", limit=limit
        )

    @result_round_digit(4)
    @source_round_digit()
    @fix_zero_division
    def calc_wind_scale_stronger_ratio(
        self,
        observation: Union[Number, np.ndarray] = None,
        forecast: Union[Number, np.ndarray] = None,
    ):
        """Calculate wind scale stronger ratio.

        Args:
            observation (Union[Number, np.ndarray]): Observation wind speed value or ndarray in m/s.
            forecast (Union[Number, np.ndarray]): Forecast wind speed value or ndarray in m/s.

        Returns:
            float: Wind scale stronger ratio(%).
        """
        if observation is None:
            observation = self.obs_spd
        if forecast is None:
            forecast = self.fct_spd

        obs_lev = identify_wind_scale(observation)
        fct_lev = identify_wind_scale(forecast)

        lev_defls = fct_lev - obs_lev

        stronger_series = np.full_like(lev_defls, False)
        stronger_series[lev_defls > 0] = True

        return Counter(stronger_series)[True] / len(stronger_series) * 100

    @result_round_digit(4)
    @source_round_digit()
    @fix_zero_division
    def calc_wind_scale_weaker_ratio(
        self,
        observation: Union[Number, np.ndarray] = None,
        forecast: Union[Number, np.ndarray] = None,
    ):
        """Calculate wind scale weaker ratio.

        Args:
            observation (Union[Number, np.ndarray]): Observation wind speed value or ndarray in m/s.
            forecast (Union[Number, np.ndarray]): Forecast wind speed value or ndarray in m/s.

        Returns:
            float: Wind scale weaker ratio(%).
        """
        if observation is None:
            observation = self.obs_spd
        if forecast is None:
            forecast = self.fct_spd

        obs_lev = identify_wind_scale(observation)
        fct_lev = identify_wind_scale(forecast)

        lev_defls = fct_lev - obs_lev

        stronger_series = np.full_like(lev_defls, False)
        stronger_series[lev_defls < 0] = True

        return Counter(stronger_series)[True] / len(stronger_series) * 100

    @result_round_digit(4)
    @source_round_digit()
    @fix_zero_division
    def calc_speed_score(
        self,
        observation: Union[Number, np.ndarray] = None,
        forecast: Union[Number, np.ndarray] = None,
    ):
        """Calculate wind speed score.

        Args:
            observation (Union[Number, np.ndarray]): Observation wind speed value or ndarray in m/s.
            forecast (Union[Number, np.ndarray]): Forecast wind speed value or ndarray in m/s.

        Returns:
            float: Wind speed score.
        """
        if observation is None:
            observation = self.obs_spd
        if forecast is None:
            forecast = self.fct_spd

        obs_lev = identify_wind_scale(observation)
        fct_lev = identify_wind_scale(forecast)

        lev_defls = get_least_lev_diff(obs_lev, fct_lev)
        score_series = np.full_like(lev_defls, 0).astype(float)

        score_series[np.isclose(lev_defls, 0)] = 1
        score_series[np.isclose(lev_defls, 1)] = 0.6
        score_series[np.isclose(lev_defls, 2)] = 0.4

        return sum(score_series) / len(score_series)

    def gather_all_factors(self):
        funcs = {
            # tuple: (function object, unit, comment)
            "direction_accuracy_ratio": (self.calc_dir_accuracy_ratio, "%", None),
            "direction_accuracy_ratio_8": (
                partial(self.calc_dir_accuracy_ratio, mode="drange8"),
                "%",
                None,
            ),
            "direction_score": (self.calc_dir_score, None, None),
            "speed_accuracy_ratio": (
                self.calc_speed_accuracy_ratio,
                "%",
                "≤2m/s",
            ),
            "wind_scale_stronger_ratio": (
                self.calc_wind_scale_stronger_ratio,
                "%",
                None,
            ),
            "wind_scale_weaker_ratio": (self.calc_wind_scale_weaker_ratio, "%", None),
            "speed_mae": (self.calc_mae, "m/s", None),
            "speed_score": (self.calc_speed_score, None, None),
            "rmse": (self.calc_rmse, "m/s", "RMSE"),
            "mae": (self.calc_mae, "m/s", "MAE"),
            "rss": (self.calc_rss, "(m/s) ** 2", None),
            "chi_square": (self.calc_chi_square, "(m/s) ** 2", None),
        }

        result = []

        for func_desc, (func, units, comment) in funcs.items():
            result.append(
                {
                    "element": "wind",
                    "interval": np.nan,
                    "kind": np.nan,
                    "level": np.nan,
                    "indicator": func_desc,
                    "value": func(),
                    "unit": units,
                    "comment": comment,
                }
            )

        slope, intercept, *_ = self.calc_linregress_args()
        if intercept > 0:
            formula = f"y = {slope} * x + {abs(intercept)}"
        else:
            formula = f"y = {slope} * x - {abs(intercept)}"
        result.append(
            {
                "element": "wind",
                "interval": np.nan,
                "kind": np.nan,
                "level": np.nan,
                "indicator": "equation_of_linear_regression",
                "value": formula,
                "unit": None,
                "comment": None,
            }
        )

        df = pd.DataFrame(result)

        return df


if __name__ == "__main__":
    pass



from .accuracy_ratio import ACCURACY_RATE_CASE
from .hit_ratio import HIT_RATIO_CASE
from .bias_score import BIAS_SCORE_CASE
from .ets import ETS_SCORE_CASE
from .false_alarm import FALSE_ALARM_RATE_CASE, FALSE_ALARM_RATIO_CASE
from .miss_ratio import MISS_RATE_CASE
from .ts import TS_SCORE_CASE



import numpy as np

from cyeva import TemperatureComparison
from .case.temp import (
    ACCURACY_RATE_CASE,
    RMSE_CASE,
    MAE_CASE,
    RESIDUAL_SUM_OF_SQUARE_CASE,
    CHI_SQUARE_CASE,
    LINREGRESS_CASE,
)


def test_calc_temp_accuracy_ratio():
    for limit, cases in ACCURACY_RATE_CASE.items():
        for case in cases:
            obs = case["obs"]
            fcst = case["fct"]
            result = case["result"]

            temp = TemperatureComparison(obs, fcst)

            _result = temp.calc_diff_accuracy_ratio(limit=limit)

            if not np.isnan(result):
                assert _result == result
            else:
                assert np.isnan(_result)


def test_calc_temp_rmse():
    for case in RMSE_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        temp = TemperatureComparison(obs, fcst)

        _result = temp.calc_rmse()

        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)


def test_calc_temp_mae():
    for case in MAE_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        temp = TemperatureComparison(obs, fcst)

        _result = temp.calc_mae()

        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)


def test_calc_temp_rss():
    for case in RESIDUAL_SUM_OF_SQUARE_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        temp = TemperatureComparison(obs, fcst)

        _result = temp.calc_rss()

        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)


def test_calc_temp_chi_square():
    for case in CHI_SQUARE_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        temp = TemperatureComparison(obs, fcst)

        _result = temp.calc_chi_square()

        if not np.isnan(result):
            assert _result == result
        else:
            assert np.isnan(_result)


def test_calc_temp_linregress():
    for case in LINREGRESS_CASE:
        obs = case["obs"]
        fcst = case["fct"]
        result = case["result"]

        temp = TemperatureComparison(obs, fcst)

        slope, intercept, *_ = temp.calc_linregress_args(obs, fcst)

        assert slope == result["slope"]
        assert intercept == result["intercept"]


def test_gather_all_factors():
    for _, cases in ACCURACY_RATE_CASE.items():
        for case in cases:
            obs = case["obs"]
            fcst = case["fct"]

            TemperatureComparison(obs, fcst)



import setuptools
import os
import codecs


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r", encoding="utf8") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


FILE_PATH = os.path.dirname(os.path.realpath(__file__))

with open(os.path.join(FILE_PATH, "README.md"), "r", encoding="utf8") as fh:
    long_description = fh.read()

requirements_path = os.path.join(FILE_PATH, "requirements/requirements.txt")
with open(requirements_path, encoding="utf8") as f:
    required = f.read().splitlines()

setuptools.setup(
    name="cyeva",
    author="caiyunapp",
    version=get_version("cyeva/__init__.py"),
    author_email="oss@caiyunapp.com",
    description="A package to evaluate weather forecast correction",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/caiyunapp/cyeva",
    include_package_data=True,
    package_data={"": ["*.csv", "*.config", "*.nl", "*.json"]},
    packages=setuptools.find_packages(),
    install_requires=required,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
        "Topic :: Scientific/Engineering :: Mathematics",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Framework :: Sphinx",
    ],
    python_requires=">=3.7",
)
